
from __future__ import absolute_import

import unittest

from pysasl import (SASLAuth, UnexpectedAuthChallenge,
                    AuthenticationCredentials)
from pysasl.oauth import OAuth2Mechanism


class TestOAuth2Mechanism(unittest.TestCase):

    def setUp(self):
        self.mech = OAuth2Mechanism()

    def test_availability(self):
        sasl = SASLAuth()
        self.assertIsNone(sasl.get(b'XOAUTH2'))
        sasl = SASLAuth([b'XOAUTH2'])
        self.assertIsInstance(sasl.get(b'XOAUTH2'), OAuth2Mechanism)
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([], sasl.server_mechanisms)
        self.assertEqual(self.mech, sasl.get(b'XOAUTH2'))

    def test_client_attempt(self):
        creds = AuthenticationCredentials('testuser', 'testtoken')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'user=testuser\x01auth=Bearertesttoken\x01\x01',
                         resp1.get_response())
        resp1.set_challenge(b'')
        resp2 = self.mech.client_attempt(creds, [resp1])
        self.assertEqual(b'', resp2.get_response())
        self.assertRaises(UnexpectedAuthChallenge,
                          self.mech.client_attempt,
                          creds, [resp1, resp2])




from __future__ import absolute_import

import unittest

from pysasl import (SASLAuth, ServerChallenge, UnexpectedAuthChallenge,
                    AuthenticationCredentials)
from pysasl.external import ExternalMechanism


class TestExternalMechanism(unittest.TestCase):

    def setUp(self):
        self.mech = ExternalMechanism()

    def test_availability(self):
        sasl = SASLAuth()
        self.assertIsNone(sasl.get(b'EXTERNAL'))
        sasl = SASLAuth([b'EXTERNAL'])
        self.assertIsInstance(sasl.get(b'EXTERNAL'), ExternalMechanism)
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)
        self.assertEqual(self.mech, sasl.get(b'EXTERNAL'))

    def test_server_attempt_issues_challenge(self):
        try:
            self.mech.server_attempt([])
        except ServerChallenge as exc:
            self.assertEqual(b'', exc.challenge)
        else:
            self.fail('ServerChallenge not raised')

    def test_server_attempt_successful(self):
        resp = ServerChallenge(b'')
        resp.set_response(b'abcdefghi')
        result, final = self.mech.server_attempt([resp])
        self.assertIsNone(final)
        self.assertFalse(result.has_secret)
        self.assertEqual('abcdefghi', result.authzid)
        self.assertEqual('abcdefghi', result.authcid)
        self.assertEqual('abcdefghi', result.identity)
        self.assertRaises(AttributeError, getattr, result, 'secret')
        self.assertTrue(result.check_secret('secret'))
        self.assertTrue(result.check_secret('invalid'))

    def test_server_attempt_successful_empty(self):
        resp = ServerChallenge(b'')
        resp.set_response(b'')
        result, _ = self.mech.server_attempt([resp])
        self.assertIsNone(result.authzid)
        self.assertEqual('', result.authcid)

    def test_client_attempt(self):
        creds = AuthenticationCredentials('', '', 'testzid')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'testzid', resp1.get_response())
        resp2 = self.mech.client_attempt(creds, [resp1])
        self.assertEqual(b'testzid', resp1.get_response())
        self.assertRaises(UnexpectedAuthChallenge,
                          self.mech.client_attempt,
                          creds, [resp1, resp2])



from __future__ import absolute_import

from . import (ServerMechanism, ClientMechanism, ServerChallenge,
               ClientResponse, AuthenticationCredentials,
               UnexpectedAuthChallenge)

__all__ = ['ExternalResult', 'ExternalMechanism']


class ExternalResult(AuthenticationCredentials):
    """Because this mechanism does not use authentication identity or secret
    strings, the :meth:`~ExternalMechanism.server_attempt` method returns this
    sub-class which only allows the :attr:`.authzid` attribute.

    """

    def __init__(self, authzid=None):
        super(ExternalResult, self).__init__('', '', authzid)

    @property
    def has_secret(self):
        return False

    @property
    def authcid(self):
        """The authentication identity string is an alias of the
        :attr:`.authzid` string for this mechanism, except it will return an
        empty string instead of ``None``.

        """
        return self.authzid or ''

    @property
    def secret(self):
        """The secret string is not available for this mechanism.

        Raises:
            AttributeError

        """
        raise AttributeError('secret')

    def check_secret(self, secret):
        """This method always returns True for this mechanism, unless
        overridden by a subclass to provide external enforcement rules.

        """
        return True


class ExternalMechanism(ServerMechanism, ClientMechanism):
    """Implements the EXTERNAL authentication mechanism.

    See Also:
        `RFC 4422 Appendix A <https://tools.ietf.org/html/rfc4422#appendix-A>`_

    Attributes:
        name: The SASL name for this mechanism.
        priority: Determines the sort ordering of this mechanism.
        insecure: This mechanism is considered secure for non-encrypted
            sessions.

    """

    name = b'EXTERNAL'
    priority = None
    insecure = False

    def server_attempt(self, challenges):
        if not challenges:
            raise ServerChallenge(b'')
        authzid = challenges[0].response
        authzid_str = authzid.decode('utf-8')
        return ExternalResult(authzid_str), None

    def client_attempt(self, creds, responses):
        if len(responses) > 1:
            raise UnexpectedAuthChallenge()
        authzid = (creds.authzid or '').encode('utf-8')
        return ClientResponse(authzid)



from __future__ import absolute_import

import re

from . import (ServerMechanism, ClientMechanism, ServerChallenge,
               ClientResponse, AuthenticationError, AuthenticationCredentials,
               UnexpectedAuthChallenge)

__all__ = ['PlainMechanism']


class PlainMechanism(ServerMechanism, ClientMechanism):
    """Implements the PLAIN authentication mechanism.

    Attributes:
        name: The SASL name for this mechanism.
        priority: Determines the sort ordering of this mechanism.
        insecure: This mechanism is not considered secure for non-encrypted
            sessions.

    """

    _pattern = re.compile(br'^([^\x00]*)\x00([^\x00]+)\x00([^\x00]*)$')

    name = b'PLAIN'
    priority = 1
    insecure = True

    def server_attempt(self, challenges):
        if not challenges:
            raise ServerChallenge(b'')

        response = challenges[0].response
        match = re.match(self._pattern, response)
        if not match:
            raise AuthenticationError('Invalid PLAIN response')
        zid, cid, secret = match.groups()

        cid_str = cid.decode('utf-8')
        secret_str = secret.decode('utf-8')
        zid_str = zid.decode('utf-8')
        return AuthenticationCredentials(cid_str, secret_str, zid_str), None

    def client_attempt(self, creds, responses):
        if len(responses) > 1:
            raise UnexpectedAuthChallenge()
        authzid = (creds.authzid or '').encode('utf-8')
        authcid = creds.authcid.encode('utf-8')
        secret = creds.secret.encode('utf-8')
        response = b'\0'.join((authzid, authcid, secret))
        return ClientResponse(response)



from __future__ import absolute_import

import re
import hmac
import hashlib
import email.utils

from . import (ServerMechanism, ClientMechanism, ServerChallenge,
               ClientResponse, AuthenticationError, UnexpectedAuthChallenge,
               AuthenticationCredentials)

__all__ = ['CramMD5Result', 'CramMD5Mechanism']


class CramMD5Result(AuthenticationCredentials):
    """Because this mechanism uses hash algorithms to compare secrets, the
    :meth:`~CramMD5Mechanism.server_attempt` method returns this sub-class
    which overrides the :meth:`.check_secret` method.

    Attributes:
        challenge: The challenge string issued by the server.
        digest: The digest computed by the client.

    """

    __slots__ = ['challenge', 'digest']

    def __init__(self, username, challenge, digest):
        super(CramMD5Result, self).__init__(username, '')
        self.challenge = challenge
        self.digest = digest

    @property
    def has_secret(self):
        return False

    @property
    def secret(self):
        """The secret string is not available in this mechanism.

        Raises:
            AttributeError

        """
        raise AttributeError('secret')

    def check_secret(self, secret):
        if not isinstance(secret, bytes):
            secret = secret.encode('utf-8')
        expected_hmac = hmac.new(secret, self.challenge, hashlib.md5)
        expected = expected_hmac.hexdigest().encode('ascii')
        try:
            return hmac.compare_digest(expected, self.digest)
        except AttributeError:  # pragma: no cover
            return expected == self.digest


class CramMD5Mechanism(ServerMechanism, ClientMechanism):
    """Implements the CRAM-MD5 authentication mechanism.

    Warning:
        Although secure during transport, offering this mechanism can be
        dangerous, as it can have implications about how the credentials are
        stored server-side.

    Attributes:
        name: The SASL name for this mechanism.
        priority: Determines the sort ordering of this mechanism.
        insecure: This mechanism is considered secure for non-encrypted
            sessions.

    """

    _pattern = re.compile(br'^(.*) ([^ ]+)$')

    name = b'CRAM-MD5'
    priority = 10
    insecure = False

    def server_attempt(self, challenges):
        if not challenges:
            challenge = email.utils.make_msgid().encode('utf-8')
            raise ServerChallenge(challenge)
        challenge = challenges[0].challenge
        response = challenges[0].response

        match = re.match(self._pattern, response)
        if not match:
            raise AuthenticationError('Invalid CRAM-MD5 response')
        username, digest = match.groups()

        username_str = username.decode('utf-8')
        return CramMD5Result(username_str, challenge, digest), None

    def client_attempt(self, creds, responses):
        if len(responses) < 1:
            return ClientResponse(b'')
        elif len(responses) > 1:
            raise UnexpectedAuthChallenge()
        challenge = responses[0].challenge
        authcid = creds.authcid.encode('utf-8')
        secret = creds.secret.encode('utf-8')
        digest = hmac.new(secret, challenge, hashlib.md5).hexdigest()
        response = b' '.join((authcid, digest.encode('ascii')))
        return ClientResponse(response)




from __future__ import absolute_import

import unittest
import email.utils

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch  # type: ignore

from pysasl import (SASLAuth, ServerChallenge, AuthenticationError,
                    UnexpectedAuthChallenge, AuthenticationCredentials)
from pysasl.crammd5 import CramMD5Mechanism


class TestCramMD5Mechanism(unittest.TestCase):

    def setUp(self):
        self.mech = CramMD5Mechanism()

    def test_availability(self):
        sasl = SASLAuth()
        self.assertIsInstance(sasl.get(b'CRAM-MD5'), CramMD5Mechanism)
        sasl = SASLAuth([b'CRAM-MD5'])
        self.assertIsInstance(sasl.get(b'CRAM-MD5'), CramMD5Mechanism)
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)
        self.assertEqual(self.mech, sasl.get(b'CRAM-MD5'))

    @patch.object(email.utils, 'make_msgid')
    def test_server_attempt_issues_challenge(self, make_msgid_mock):
        make_msgid_mock.return_value = '<abc123.1234@testhost>'
        try:
            self.mech.server_attempt([])
        except ServerChallenge as exc:
            self.assertEqual(b'<abc123.1234@testhost>', exc.challenge)
        else:
            self.fail('ServerChallenge not raised')

    @patch.object(email.utils, 'make_msgid')
    def test_server_attempt_bad_response(self, make_msgid_mock):
        make_msgid_mock.return_value = '<abc123.1234@testhost>'
        resp = ServerChallenge(b'')
        resp.set_response(b'testing')
        self.assertRaises(AuthenticationError,
                          self.mech.server_attempt, [resp])

    @patch.object(email.utils, 'make_msgid')
    def test_server_attempt_successful(self, make_msgid_mock):
        make_msgid_mock.return_value = '<abc123.1234@testhost>'
        response = b'testuser 3a569c3950e95c490fd42f5d89e1ef67'
        resp = ServerChallenge(b'<abc123.1234@testhost>')
        resp.set_response(response)
        result, final = self.mech.server_attempt([resp])
        self.assertIsNone(final)
        self.assertFalse(result.has_secret)
        self.assertIsNone(result.authzid)
        self.assertEqual('testuser', result.authcid)
        self.assertEqual('testuser', result.identity)
        self.assertRaises(AttributeError, getattr, result, 'secret')
        self.assertTrue(result.check_secret(u'testpass'))
        self.assertTrue(result.check_secret(b'testpass'))
        self.assertFalse(result.check_secret('badpass'))

    def test_client_attempt(self):
        creds = AuthenticationCredentials('testuser', 'testpass')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'', resp1.get_response())
        resp1.set_challenge(b'<abc123.1234@testhost>')
        resp2 = self.mech.client_attempt(creds, [resp1])
        self.assertEqual(b'testuser 3a569c3950e95c490fd42f5d89e1ef67',
                         resp2.get_response())
        self.assertRaises(UnexpectedAuthChallenge,
                          self.mech.client_attempt,
                          creds, [resp1, resp2])




from __future__ import absolute_import

import unittest

from pysasl import (SASLAuth, ServerChallenge, AuthenticationCredentials,
                    UnexpectedAuthChallenge)
from pysasl.login import LoginMechanism


class TestLoginMechanism(unittest.TestCase):

    def setUp(self):
        self.mech = LoginMechanism()

    def test_availability(self):
        sasl = SASLAuth()
        self.assertIsInstance(sasl.get(b'LOGIN'), LoginMechanism)
        sasl = SASLAuth([b'LOGIN'])
        self.assertIsInstance(sasl.get(b'LOGIN'), LoginMechanism)
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)
        self.assertEqual(self.mech, sasl.get(b'LOGIN'))

    def test_server_attempt_issues_challenges(self):
        try:
            self.mech.server_attempt([])
        except ServerChallenge as exc:
            self.assertEqual(b'Username:', exc.challenge)
        else:
            self.fail('first ServerChallenge not raised')
        try:
            self.mech.server_attempt(['test'])
        except ServerChallenge as exc:
            self.assertEqual(b'Password:', exc.challenge)
        else:
            self.fail('second ServerChallenge not raised')

    def test_server_attempt_successful(self):
        resp1 = ServerChallenge(b'Username:')
        resp1.set_response(b'testuser')
        resp2 = ServerChallenge(b'Password:')
        resp2.set_response(b'testpass')
        result, final = self.mech.server_attempt([resp1, resp2])
        self.assertIsNone(final)
        self.assertTrue(result.has_secret)
        self.assertIsNone(result.authzid)
        self.assertEqual('testuser', result.authcid)
        self.assertEqual('testuser', result.identity)
        self.assertTrue(result.check_secret('testpass'))
        self.assertFalse(result.check_secret('invalid'))

    def test_client_attempt(self):
        creds = AuthenticationCredentials('testuser', 'testpass')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'', resp1.get_response())
        resp1.set_challenge(b'Username:')
        resp2 = self.mech.client_attempt(creds, [resp1])
        self.assertEqual(b'testuser', resp2.get_response())
        resp2.set_challenge(b'Password:')
        resp3 = self.mech.client_attempt(creds, [resp1, resp2])
        self.assertEqual(b'testpass', resp3.get_response())
        self.assertRaises(UnexpectedAuthChallenge,
                          self.mech.client_attempt,
                          creds, [resp1, resp2, resp3])



from __future__ import absolute_import

import heapq
import hmac
from collections import OrderedDict
from functools import total_ordering

from pkg_resources import iter_entry_points

__all__ = ['AuthenticationError', 'UnexpectedAuthChallenge',
           'AuthenticationCredentials', 'BaseMechanism',
           'ServerChallenge', 'ServerMechanism',
           'ClientResponse', 'ClientMechanism',
           'SASLAuth']


class AuthenticationError(Exception):
    """Indicates that authentication failed due to a protocol error unrelated
    to any provided credentials.

    """
    pass


class UnexpectedAuthChallenge(AuthenticationError):
    """During client-side authentication, the SASL mechanism received an
    authentication challenge from the server that it did not expect.

    """

    def __init__(self):
        msg = 'Unexpected auth challenge.'
        super(UnexpectedAuthChallenge, self).__init__(msg)


class AuthenticationCredentials(object):
    """Object returned by :meth:`~ServerMechanism.server_attempt` and passed in
    to :meth:`~ClientMechanism.client_attempt` containing information about the
    authentication credentials in use.

    Args:
        authcid: Authentication ID string (the username).
        secret: Secret string (the password).
        authzid: Authorization ID string, if applicable.

    """

    __slots__ = ['_authcid', '_secret', '_authzid']

    def __init__(self, authcid, secret, authzid=None):
        super(AuthenticationCredentials, self).__init__()
        self._authcid = authcid
        self._secret = secret
        self._authzid = authzid or None

    @property
    def has_secret(self):
        """True if the :attr:`.secret` attribute is valid for this credentials
        type.

        If this returns False, the :attr:`.secret` attribute should raise
        :exc:`AttributeError`.

        """
        return True

    @property
    def authcid(self):
        """The authentication identity string used in the attempt."""
        return self._authcid

    @property
    def secret(self):
        """Contains the secret string used in the authentication attempt,
        if available. Use :meth:`.check_secret` instead, when possible.

        """
        return self._secret

    @property
    def authzid(self):
        """The authorization identity string used in the attempt, or ``None``
        if this field is empty or unused.

        """
        return self._authzid

    @property
    def identity(self):
        """The canonical identity being assumed by the authentication attempt.
        This is :attr:`.authzid` if available, :attr:`.authcid` otherwise.

        Consider a UNIX system where ``root`` is the superuser and only it may
        assume the identity of other users. With an :attr:`.authcid` of
        ``root`` and an :attr:`.authzid` of ``terry``, the authorization would
        succeed and :attr:`.identity` would be ``terry``. With an
        :attr:`.authcid` of ``greg``, authorization would fail because ``greg``
        is not the superuser and cannot assume the :attr:`.identity` of
        ``terry``.

        """
        if self.authzid is not None:
            return self.authzid
        else:
            return self.authcid

    def check_secret(self, secret):
        """Checks if the secret string used in the authentication attempt
        matches the "known" secret string. Some mechanisms will override this
        method to control how this comparison is made.

        Args:
            secret: The secret string to compare against what was used in the
                authentication attempt.

        Returns:
            True if the given secret matches the authentication attempt.

        """
        if isinstance(secret, bytes):
            secret = secret.decode('utf-8')
        try:
            return hmac.compare_digest(secret, self.secret)
        except AttributeError:  # pragma: no cover
            return secret == self.secret


class ClientResponse(object):
    """Used by :meth:`~ClientMechanism.client_attempt` to provide client
    responses and to populate server challenges.

    Args:
        response: The response string that should be sent to the server.

    """

    __slots__ = ['response', 'challenge']

    def __init__(self, response):
        super(ClientResponse, self).__init__()
        self.response = response
        self.challenge = None

    def get_response(self):
        """Return the client response that should be sent to the server."""
        return self.response

    def set_challenge(self, data):
        """If the server reacts to the response with a challenge, set it with
        this method.

        Args:
            data: The challenge string.

        """
        self.challenge = data


class ServerChallenge(Exception):
    """Raised by :meth:`~ServerMechanism.server_attempt` to provide server
    challenges and to populate client responses.

    Args:
        challenge: The challenge string that should be sent to the client.

    """

    __slots__ = ['challenge', 'response']

    def __init__(self, challenge):
        super(ServerChallenge, self).__init__()
        self.challenge = challenge
        self.response = None

    def get_challenge(self):
        """Return the server challenge that should be sent to the client."""
        return self.challenge

    def set_response(self, data):
        """After the challenge is sent to the client, its response should be
        set with this method.

        Args:
            data: The response string.

        """
        self.response = data


@total_ordering
class BaseMechanism(object):
    """Base class for all server- and client-side SASL mechanisms.

    Attributes:
        name: The SASL name for this mechanism.
        priority: Determines the sort ordering of this mechanism.
        insecure: Whether this mechanism is considered secure for non-encrypted
            sessions. This value should be used by implementations to determine
            which mechanisms are chosen or offered.

    """

    name = b''
    priority = None
    insecure = False

    def __lt__(self, other):
        if not isinstance(other, BaseMechanism):
            return NotImplemented
        elif other.priority is None:
            return False
        elif self.priority is None:
            return True
        else:
            return self.priority < other.priority


class ServerMechanism(BaseMechanism):
    """Base class for implementing SASL mechanisms that support server-side
    credential verification.

    """

    def server_attempt(self, challenges):  # pragma: no cover
        """For SASL server-side credential verification, receives responses
        from the client and issues challenges until it has everything needed to
        verify the credentials.

        If a challenge is necessary, an :class:`ServerChallenge` exception will
        be raised. Send the challenge string to the client with
        :meth:`~ServerChallenge.get_challenge` and then populate its response
        with :meth:`~ServerChallenge.set_response`. Finally, append the
        exception to the ``challenges`` argument before calling again.

        Args:
            challenges: The server challenges that have been issued by the
                mechanism and responded to by the client.

        Returns:
            A tuple of the authentication credentials received from the client
            once no more challenges are necessary, and an optional final
            response string from the server used by some mechanisms.

        Raises:
            ServerChallenge: The server challenge needing a client response.

        """
        raise NotImplementedError()


class ClientMechanism(BaseMechanism):
    """Base class for implementing SASL mechanisms that support client-side
    credential verification.

    """

    def client_attempt(self, creds, responses):  # pragma: no cover
        """For SASL client-side credential verification, produce responses to
        send to the server and react to its challenges until the server returns
        a final success or failure.

        Send the response string to the server with the
        :meth:`~ClientResponse.get_response` method of the returned
        :class:`ClientResponse` object. If the server returns another
        challenge, set it with the object's
        :meth:`~ClientResponse.set_challenge` method and append the object to
        the ``responses`` argument before calling again.

        Args:
            creds: The credentials to attempt authentication with.
            responses: The client responses that have been sent to the server.
                New attempts begin with an empty list.

        Returns:
            The client response to the most recent server challenge.

        Raises:
            UnexpectedAuthChallenge: The server has issued an unexpected
                challenge the client mechanism does not recognize.

        """
        raise NotImplementedError()


class SASLAuth(object):
    """Manages the mechanisms available for authentication attempts.

    Args:
        advertised: List of available SASL mechanism objects. Using the name
            of a built-in mechanism (e.g. ``b'PLAIN'``) works as well. By
            default, all built-in mechanisms are available.

    """

    __slots__ = ['mechs']

    _builtin_mechanisms = None

    def __init__(self, advertised=None):
        super(SASLAuth, self).__init__()
        if not advertised:
            builtin_mechs = self._get_builtin_mechanisms()
            advertised = [mech for mech in builtin_mechs.values()
                          if mech.priority is not None]
        self.mechs = OrderedDict()
        for mech in advertised:
            if isinstance(mech, BaseMechanism):
                self.mechs[mech.name] = mech
            else:
                builtin_mechs = self._get_builtin_mechanisms()
                self.mechs[mech] = builtin_mechs[mech]

    @classmethod
    def secure(cls):
        """Uses only authentication mechanisms that are secure for use in
        non-encrypted sessions.

        Returns:
            A new :class:`SASLAuth` object.

        """
        builtin_mechs = cls._get_builtin_mechanisms()
        secure_mechs = [mech for _, mech in builtin_mechs.items()
                        if not mech.insecure and mech.priority is not None]
        return SASLAuth(secure_mechs)

    @classmethod
    def _get_builtin_mechanisms(cls):
        if cls._builtin_mechanisms is None:
            heap = []
            mechs = OrderedDict()
            for entry_point in iter_entry_points('pysasl.mechanisms'):
                mech_cls = entry_point.load()
                heapq.heappush(heap, mech_cls())
            for i in range(len(heap)):
                mech = heapq.heappop(heap)
                mechs[mech.name] = mech
            cls._builtin_mechanisms = mechs
        return cls._builtin_mechanisms

    @property
    def server_mechanisms(self):
        """List of available :class:`ServerMechanism` objects."""
        return [mech for mech in self.mechs.values()
                if isinstance(mech, ServerMechanism)]

    @property
    def client_mechanisms(self):
        """List of available :class:`ClientMechanism` objects."""
        return [mech for mech in self.mechs.values()
                if isinstance(mech, ClientMechanism)]

    def get(self, name):
        """Get a SASL mechanism by name. The resulting object should inherit
        either :class:`ServerMechanism`, :class:`ClientMechanism`, or both.

        Args:
            name: The SASL mechanism name.

        Returns:
            The mechanism object or ``None``

        """
        return self.mechs.get(name.upper())

    def get_server(self, name):
        """Like :meth:`.get`, but only mechanisms inheriting
        :class:`ServerMechanism` will be returned.

        Args:
            name: The SASL mechanism name.

        Returns:
            The mechanism object or ``None``

        """
        mech = self.get(name)
        return mech if isinstance(mech, ServerMechanism) else None

    def get_client(self, name):
        """Like :meth:`.get`, but only mechanisms inheriting
        :class:`ClientMechanism` will be returned.

        Args:
            name: The SASL mechanism name.

        Returns:
            The mechanism object or ``None``

        """
        mech = self.get(name)
        return mech if isinstance(mech, ClientMechanism) else None



# Copyright (c) 2019 Ian C. Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from setuptools import setup, find_packages  # type: ignore


with open('README.md', 'r') as fh:
    long_description = fh.read()


setup(name='pysasl',
      version='0.4.0',
      author='Ian Good',
      author_email='icgood@gmail.com',
      description='Pure Python SASL client and server library.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='MIT',
      url='https://github.com/icgood/pysasl/',
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
      packages=find_packages(),
      package_data={'pysasl': ['py.typed']},
      entry_points={'pysasl.mechanisms': [
          'PLAIN = pysasl.plain:PlainMechanism',
          'LOGIN = pysasl.login:LoginMechanism',
          'CRAM-MD5 = pysasl.crammd5:CramMD5Mechanism',
          'EXTERNAL = pysasl.external:ExternalMechanism',
          'XOAUTH2 = pysasl.oauth:OAuth2Mechanism',
      ]},
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7'])



from __future__ import absolute_import

from . import (ServerMechanism, ClientMechanism, ServerChallenge,
               ClientResponse, AuthenticationCredentials,
               UnexpectedAuthChallenge)

__all__ = ['LoginMechanism']


class LoginMechanism(ServerMechanism, ClientMechanism):
    """Implements the LOGIN authentication mechanism.

    Attributes:
        name: The SASL name for this mechanism.
        priority: Determines the sort ordering of this mechanism.
        insecure: This mechanism is not considered secure for non-encrypted
            sessions.

    """

    name = b'LOGIN'
    priority = 5
    insecure = True

    def server_attempt(self, challenges):
        if len(challenges) < 1:
            raise ServerChallenge(b'Username:')
        if len(challenges) < 2:
            raise ServerChallenge(b'Password:')
        username = challenges[0].response.decode('utf-8')
        password = challenges[1].response.decode('utf-8')
        return AuthenticationCredentials(username, password), None

    def client_attempt(self, creds, responses):
        if len(responses) < 1:
            return ClientResponse(b'')
        if len(responses) < 2:
            username = creds.authcid.encode('utf-8')
            return ClientResponse(username)
        if len(responses) < 3:
            password = creds.secret.encode('utf-8')
            return ClientResponse(password)
        raise UnexpectedAuthChallenge()



from __future__ import absolute_import

from . import (ClientMechanism, ClientResponse, UnexpectedAuthChallenge)

__all__ = ['OAuth2Mechanism']


class OAuth2Mechanism(ClientMechanism):
    """Implements the `XOAUTH2`_ authentication mechanism, used by `Oauth 2.0`_
    systems to authenticate using access tokens.

    This mechanism is only available for client-side authentication.

    Attributes:
        name: The SASL name for this mechanism.
        priority: Determines the sort ordering of this mechanism.
        insecure: This mechanism is considered secure for non-encrypted
            sessions.

    .. _XOAUTH2: https://developers.google.com/gmail/xoauth2_protocol
    .. _OAuth 2.0: http://tools.ietf.org/html/draft-ietf-oauth-v2-22

    """

    name = b'XOAUTH2'
    priority = None
    insecure = False

    def client_attempt(self, creds, responses):
        if len(responses) > 1:
            raise UnexpectedAuthChallenge()
        elif len(responses) > 0:
            return ClientResponse(b'')
        user = creds.authcid.encode('utf-8')
        token = creds.secret.encode('utf-8')
        response = b''.join((b'user=', user, b'\x01auth=Bearer', token,
                             b'\x01\x01'))
        return ClientResponse(response)
